#!/usr/bin/python

import sys
import os
import time
import pickle
import shutil

from poaupdater import uTasks, uSysDB, uLogging, uUtil, uAction, uHCL

import config
from config import MigrationError, PrerequisiteError
#
#	Utilities
#

#	Utility functions


def execCommand(command):
    out, err, status = uUtil.readCmdExt(command)

    if status > 0:
        raise MigrationError("%s exited with non-zero status %s. (%s)" % (command, status, err))
    elif status < 0:
        raise MigrationError("%s is terminated by signal %s. (%s)" % (command, status, err))

    return status


def normalizeIP(ipstr):
    parts = tuple([int(part) for part in ipstr.split('.')])
    assert 4 == len(parts)

    return "%03d.%03d.%03d.%03d" % parts


def isIpv4(ipstr):
    return 4 == len(ipstr.split('.'))


def pause():
    print("PAUSE: Press ^D to continue")
    sys.stdin.read()


def pausable(func):
    def tmpfunc(*args, **kwargs):
        func(*args, **kwargs)
        if config.config.demomode:
            pause()
    return tmpfunc


def beautyRetriable(func):
    def tmpfunc(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except uUtil.ExecFailed, e:
            uLogging.err("%s exited with non-zero status %s", e.command, e.status)
            uLogging.err("%s", e)

            raise MigrationError("Remote command execution failed. Probably host is down.")

    return uAction.retriable(tmpfunc)


def defaultWritablePath(filename):
    # make migration script self contained - put all the logs and checkpoints into dir with main script
    MODULE_DIR = os.path.dirname(os.path.realpath(__file__))
    return os.path.abspath(os.path.join(MODULE_DIR, "../", filename))


def waitTasksComplete(where):
    con = uSysDB.connect()
    cur = con.cursor()
    prev_num = 0
    while True:
        query = "SELECT t.name, t.task_id, t.status FROM tm_tasks t WHERE (" + \
            where + ") AND t.status not in ('c') ORDER BY task_id"
        cur.execute(query)
        tasks = [(row[0], row[1], row[2]) for row in cur.fetchall()]
        if not tasks:
            return
        failed = [t for t in tasks if t[2] not in ('s', 'u', 'e')]

        if failed:
            for t in failed:
                name, tid, status = t
                cur2 = con.cursor()
                cur2.execute("SELECT action_output FROM tm_logs WHERE task_id = %s ORDER BY finished DESC", tid)
                row = cur2.fetchone()
                if row:
                    output = str(row[0])
                else:
                    output = ' no output'
                uLogging.err("\"%s\"(id=%s) failed with \"%s\"", name, tid, output)
            raise Exception("There are failed update tasks")
        if prev_num != len(tasks):
            # nm, tid, status = tasks[0]
            running = ', '.join([t[0] for t in tasks if t[2] == 'e'])

            uLogging.debug("%s (%s more to go)", running, len(tasks))
        else:
            sys.stdout.write('.')
            sys.stdout.flush()
        prev_num = len(tasks)
        time.sleep(0.1)


def countTasks(where):
    con = uSysDB.connect()
    cur = con.cursor()

    query = "SELECT COUNT(*) FROM tm_tasks t WHERE (" + where + ") AND t.status not in ('c', 'd')"
    cur.execute(query)
    return cur.fetchone()[0]


def waitTasksCompleteRef(pobjref):
    waitTasksComplete("t.location = '%s'" % ":".join(pobjref.to_db()))


def waitChangeServiceTemplate():
    uLogging.info(
        "Waiting for Apache restart. To speed things up you may run tasks 'Restart service Apache' and 'Dump config image for domain service' manually from POA task manager.")
    # wait for scheduled tasks to complete and properly update subscriptions
    beautyRetriable(waitTasksComplete)(
        "name ilike 'Dump config image for domain service%' or name ilike 'Restart service Apache%'")


def countChangeServiceTemplateTasks():
    return countTasks("name ilike 'Dump config image for domain service%' or name ilike 'Restart service Apache%'")


def waitProFTPDDumpConfig():
    uLogging.info(
        "Waiting for FTP changes to propagate. Processing 'Dump virtual host config for ProFTPD service' tasks.")
    # wait for scheduled tasks to complete and properly update subscriptions
    beautyRetriable(waitTasksComplete)("name ilike 'Dump virtual host config%'")


def waitWebAppDeletion():
    uLogging.info("Waiting for WebApps to be removed. Processing 'Deletion of WebApp from domain with ds_id=xx' tasks.")
    # wait for scheduled tasks to complete and properly update subscriptions
    beautyRetriable(waitTasksComplete)("name ilike 'Deletion of%from domain with ds_id=%'")

#
# 	Check points, ability to save migration state
#
#
CHECKPOINT_INITIAL = 0
CHECKPOINT_BEFORE_COMMIT = 1
CHECKPOINT_AFTER_COMMIT = 2
CHECKPOINT_HOST_SHUTDOWN = 3
CHECKPOINT_AFTER_RESTORE_IP_TYPES = 3


class Checkpoint:
    CURRENT_VERSION = 7

    def __load(filepath):
        f = open(filepath)
        result = pickle.load(f)
        f.close()

        return result

    load = staticmethod(__load)

    def __save(point, checkpoint, filepath=None):
        if not filepath:
            filepath = config.config.checkpointfile

        tmppath = filepath + ".tmp"

        f = open(tmppath, "w")
        checkpoint.point = point
        pickle.dump(checkpoint, f)
        f.close()

        shutil.move(tmppath, filepath)

    save = staticmethod(__save)

    def __init__(self, host, cluster, subs, config):
        self.version = Checkpoint.CURRENT_VERSION
        if host:
            self.hostID = host.hostID
        else:
            self.hostID = None
        self.clusterID = cluster.hostID
        self.subs = subs
        self.config = config
        self.point = CHECKPOINT_INITIAL

        self.mainProftpdVhostID = 0

    def __str__(self):
        return "Checkpoint(host:%s, cluster:%s, point:%s): subs:%s, config:%s, mainProftpdVhostID:%s" % (self.hostID, self.clusterID, self.point, self.subs, self.config, self.mainProftpdVhostID)

    def __setstate__(self, state):
        # do not allow loading of old checkpoints
        if not "version" in state:
            version = None
        else:
            version = state["version"]

        if not version or version < Checkpoint.CURRENT_VERSION:
            raise MigrationError("Old version %s of checkpoint cannot be loaded. Current checkpoint version is %s." % (
                version, Checkpoint.CURRENT_VERSION))

        self.__dict__.update(state)

#
#	Manage POA customer ssh access
#


class SSHAccess:

    def fetchParams(self, cur, webspace):
        cur.execute("SELECT login, password, state FROM apache_ssh_access WHERE homedir_id = %s", webspace.webspaceID)
        rv = cur.fetchone()
        if not rv:
            return ("", "", "")

        return (rv[0], rv[1], rv[2])

    def createTask(self, name, method, webspace, host):
        pobjref = uTasks.PObjRef(uTasks.SCRef("apache"), "getDomainService", webspace.getDomainServiceID("Apache"))

        task = uTasks.Task(pobjref, name, method)
        task.setParam("service_id", host.getServiceID("Apache"))
        task.setParam("webspace_id", str(webspace.webspaceID))

        return (pobjref, task)

    def enable(self, webspace):
        con = uSysDB.connect()
        cur = con.cursor()

        login, password, state = self.fetchParams(cur, webspace)
        if not login:
            return

        if 'e' == state:
            uLogging.debug("SSH access for webspace %s is already enabled", webspace.webspaceID)
            return

        cur.execute(
            "UPDATE apache_ssh_access SET state = 'e', ops_scheduled = ops_scheduled + 1 WHERE homedir_id = %s", webspace.webspaceID)
        cur.close()
        con.commit()

        pobjref, task = self.createTask("Enable SSH access for webspace %s" %
                                        webspace.webspaceID, "taskEnableSSHAccess", webspace, webspace.host)
        task.setParam("login", login)
        task.setParam("password", password)
        task.schedule()

    def disable(self, webspace):
        con = uSysDB.connect()
        cur = con.cursor()

        login, password, state = self.fetchParams(cur, webspace)
        if not login or "d" == state:
            return

        cur.execute(
            "UPDATE apache_ssh_access SET state = 'd', ops_scheduled = ops_scheduled + 1 WHERE homedir_id = %s", webspace.webspaceID)
        cur.close()
        con.commit()

        pobjref, task = self.createTask("Disable SSH access for webspace %s" %
                                        webspace.webspaceID, "taskDisableSSHAccess", webspace, webspace.host)
        task.setParam("lock", "1")
        task.setParam("login", login)
        task.schedule()

        beautyRetriable(waitTasksCompleteRef)(pobjref)

#
#	Extended hcl request class, to support operations which didn't exist in 5.2
#


class HCLRequest(uHCL.Request):

    def __init__(self, host_id=None, user=None, group=None, auto_export=True):
        uHCL.Request.__init__(self, host_id, user, group, auto_export)

    def ipadd(self, ip_address, netmask, iface):
        elem = self._Request__document.createElement("IPADD")
        elem.setAttribute("ip", ip_address)
        elem.setAttribute("netmask", netmask)
        elem.setAttribute("ifname", iface)

        self._Request__perform.appendChild(elem)

    def ipdel(self, ip_address, netmask, iface):
        elem = self._Request__document.createElement("IPDEL")
        elem.setAttribute("ip", ip_address)
        elem.setAttribute("netmask", netmask)
        elem.setAttribute("ifname", iface)

        self._Request__perform.appendChild(elem)


def batchExecute(objects, func, batchSize=100):
    index = 0
    objectsLength = len(objects)
    while index < objectsLength:
        batch = objects[index:index + batchSize]
        uLogging.info("Batch %s\%s" % (index / batchSize + 1, objectsLength / batchSize + 1))
        func(batch)
        index += batchSize


def isNumber(x):
    try:
        int(x)
        return True
    except:
        return False
